import { dirname } from "https://deno.land/std@0.224.0/path/dirname.ts";
import { exists } from "https://deno.land/std@0.224.0/fs/exists.ts";
import { assertExists } from "https://deno.land/std@0.224.0/assert/assert_exists.ts";

export function* range(from: number, to: number) {
	while (from < to) yield from++;
}

export async function project_root() {
	const this_url = new URL(import.meta.url);
	const this_abs_path = await Deno.realPath(this_url.pathname);
	return dirname(dirname(dirname(this_abs_path)));
}

export async function wait(ms: number) {
	await new Promise((res) => setTimeout(res, ms));
}

export async function wait_file(path: string, delay_ms: number = 10) {
	while (true) {
		if (await exists(path)) return;
		await wait(delay_ms);
	}
}

export type Consumer<T> = (value: T) => void;

export function split_promise<T>() {
	let resovler = null as Consumer<T> | null;
	const promise = new Promise<T>((res) => resovler = res);
	assertExists(resovler);
	return [promise, resovler] as const;
}

export class Lock<T> {
	private value;
	private waiting_queue;
	private in_use;

	public constructor(value: T) {
		this.value = value;
		this.waiting_queue = [] as Consumer<void>[];
		this.in_use = new Ref(false);
	}

	public async with<O>(procedure: (guard: Guard<T>) => Promise<O>) {
		const guard = await this.lock();
		const result = await procedure(guard);
		guard.release();
		return result;
	}

	public async lock() {
		if (this.in_use.value) {
			const [promise, consumer] = split_promise<void>();
			this.waiting_queue.push(consumer);
			await promise;
		}
		this.in_use.value = true;
		return new Guard(this.value, this.waiting_queue, this.in_use);
	}
}

export class Guard<T> {
	private value;
	private waiting_queue;
	private in_use;

	constructor(
		value: T,
		waiting_queue: Consumer<void>[],
		in_use: Ref<boolean>,
	) {
		this.value = value;
		this.waiting_queue = waiting_queue;
		this.in_use = in_use;
	}

	public release() {
		this.in_use.value = false;
		const [first] = this.waiting_queue.splice(0, 1);
		if (first !== undefined) first();
	}

	public get() {
		return this.value;
	}
}

export class Ref<T> {
	public value;

	public constructor(value: T) {
		this.value = value;
	}
}
